﻿/*
*   This file is part of NSMB Editor 5.
*
*   NSMB Editor 5 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   NSMB Editor 5 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with NSMB Editor 5.  If not, see <http://www.gnu.org/licenses/>.
*/

﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace NSMBe4.Patcher
{
    public class PatchMaker
    {
        //DY
        private System.IO.DirectoryInfo patchDir = new System.IO.DirectoryInfo(ROM.romfile.Directory.FullName + "/source");
        private System.IO.DirectoryInfo tmpDir = new System.IO.DirectoryInfo(ROM.romfile.Directory.FullName + "/source/_tmp_DoNotUse_WillBeDeleted");

        private int ArenaLoOffs;
        Arm9BinaryHandler handler;
        DirectoryInfo romdir;

        public PatchMaker(DirectoryInfo romdir)
        {
            handler = new Arm9BinaryHandler();
            this.romdir = romdir;
        }

		public uint getCodeAddr()
		{
            //handler.load();  //DY: moved to compilePatch()
            loadArenaLoOffsFile(romdir);
            uint codeAddr = handler.readFromRamAddr(ArenaLoOffs, -1);
            return codeAddr;
		}
		
		public void restore()
		{
            handler.restoreFromBackup();
		}

        public void compilePatch()
        {
            //DY
            if (patchDir.Exists && patchDir.GetFiles().Length > 0)
            {
                // init
                handler.load();   // loads ARM9 sections, required for overwriting to ARM9
                if (tmpDir.Exists)
                    tmpDir.Delete(true);
                tmpDir.Create();

                // compile and insert standard hooks
                PatchCompiler.cleanPatch(ROM.romfile.Directory); // ensure "make" runs (otherwise doesn't run if output files are newer than source files)
                uint codeAddr = getCodeAddr();
                PatchCompiler.compilePatch(codeAddr, romdir, patchDir.Name);
                generatePatch();    // moved from LevelChooser.makeinsert_Click()

                // parse and insert overwrites
                saveBuildFiles();   // needed so that address of "new_" labels can be found
                PatchCompiler.cleanPatch(ROM.romfile.Directory); // ensure "make" runs (otherwise doesn't run if output files are newer than source files)
                parseOverwriteFiles();

                // cleanup
                handler.saveSections(); // save ARM9 sections, required to actually overwrite to ARM9
                restoreBuildFiles();    // optional - restores full initial build files
                tmpDir.Delete(true);
            }
        }

        //DY
        public void saveBuildFiles()
        {
            System.IO.File.Move(romdir.FullName + "/newcode.bin", tmpDir.FullName + "/newcode.bin");
            System.IO.File.Move(romdir.FullName + "/newcode.elf", tmpDir.FullName + "/newcode.elf");
            System.IO.File.Move(romdir.FullName + "/newcode.sym", tmpDir.FullName + "/newcode.sym");
            System.IO.Directory.Move(romdir.FullName + "/build", tmpDir.FullName + "/build");
        }

        public void restoreBuildFiles()
        {
            PatchCompiler.cleanPatch(ROM.romfile.Directory);
            System.IO.File.Move(tmpDir.FullName + "/newcode.bin", romdir.FullName + "/newcode.bin");
            System.IO.File.Move(tmpDir.FullName + "/newcode.elf", romdir.FullName + "/newcode.elf");
            System.IO.File.Move(tmpDir.FullName + "/newcode.sym", romdir.FullName + "/newcode.sym");
            System.IO.Directory.Move(tmpDir.FullName + "/build", romdir.FullName + "/build");
        }

        public void parseOverwriteFiles()
        {
            System.IO.DirectoryInfo ovrDir = new System.IO.DirectoryInfo(tmpDir.FullName + "/overwrites");
            ovrDir.Create();

            // extract, compile and insert each "ovr_" block in each file individually, so they can be compiled with correct CODEADDR
            System.IO.FileInfo[] ovrFiles = patchDir.GetFiles();
            foreach (System.IO.FileInfo srcFile in ovrFiles)
            {
                // only process ".s" files
                if (Path.GetExtension(srcFile.FullName) == ".s")
                {

                string[] srcText = System.IO.File.ReadAllLines(srcFile.FullName);
                bool insideMLC = false;

                int i = 0;
                    while (i < srcText.Length)
                    {
                        string l = srcText[i];

                        // skip everything inside multi-line comment block
                        {
                            if (insideMLC)
                            {
                                while (!l.Contains("*/"))
                                {
                                    i++;
                                    l = srcText[i];
                                }
                                insideMLC = false;
                                l = l.Remove(0, l.IndexOf("*/") + 2);   // retain anything appearing after "*/" on same line
                            }

                            while (l.Contains("/*") &&
                                !(l.Contains("/*overwrite") && l.Trim().Substring(0, 11) == "/*overwrite") &&   // special keyword - ignored by hook compiler, but extracted and compiled by overwrite parser
                                !(l.Contains("ovr_") && l.Trim().Substring(0, 4) == "ovr_"))    // make sure comment block doesn't appear after "ovr_" keyword
                            {

                                // check if multi-line comment block ends on same line
                                if (l.Contains("*/"))
                                {
                                    l = l.Remove(l.IndexOf("/*"), l.IndexOf("*/") - l.IndexOf("/*") + 2);   // remove everything between "/*" and "*/" on same line
                                }
                                else
                                {
                                    insideMLC = true;
                                    l = l.Remove(l.IndexOf("/*"), l.Length - l.IndexOf("/*"));  // retain anything appearing before "/*" on same line
                                }
                            }
                        }

                        // not start of "ovr_" block, keep looking
                        if (!(l.Contains("ovr_") && l.Trim().Substring(0, 4) == "ovr_"))
                        {
                            i++;
                        }

                        // found start of "ovr_" block
                        else
                        {
                            int codeAddr = parseHex(l.Trim().Substring(4, 8));

                            int ovId = -1;
                            if (l.Contains("_ov_"))
                                ovId = parseHex(l.Substring(l.IndexOf("_ov_") + 4, 2));

                            string destFile = System.IO.Path.Combine(ovrDir.FullName, srcFile.Name);
                            System.IO.FileStream fs = System.IO.File.Create(destFile);
                            fs.Close();

                            // keep copying lines to temp file until we reach end of file or next hook/overwrite block
                            using (System.IO.StreamWriter f = new System.IO.StreamWriter(destFile, true))
                            {
                                do
                                {
                                    // skip everything inside multi-line comment block
                                    {
                                        if (insideMLC)
                                        {
                                            while (!l.Contains("*/"))
                                            {
                                                i++;
                                                l = srcText[i];
                                            }
                                            insideMLC = false;
                                            l = l.Remove(0, l.IndexOf("*/") + 2);   // retain anything appearing after "*/" on same line
                                        }

                                        while (l.Contains("/*") &&
                                            !(l.Contains("/*overwrite") && l.Trim().Substring(0, 11) == "/*overwrite") &&   // special keyword - ignored by hook compiler, but extracted and compiled by overwrite parser
                                            !(l.Contains("ovr_") && l.Trim().Substring(0, 4) == "ovr_"))    // make sure comment block doesn't appear after "ovr_" keyword
                                        {

                                            // check if multi-line comment block ends on same line
                                            if (l.Contains("*/"))
                                            {
                                                l = l.Remove(l.IndexOf("/*"), l.IndexOf("*/") - l.IndexOf("/*") + 2);   // remove everything between "/*" and "*/" on same line
                                            }
                                            else
                                            {
                                                insideMLC = true;
                                                l = l.Remove(l.IndexOf("/*"), l.Length - l.IndexOf("/*"));  // retain anything appearing before "/*" on same line
                                            }
                                        }
                                    }

                                    // parse any "new_" label, copy line and move to next
                                    l = parseLabel(l);
                                    f.WriteLine(l);
                                    i++;
                                    if (i < srcText.Length)
                                        l = srcText[i];
                                    else
                                        break;  // end of file

                                } while (!(l.Contains("ovr_") && l.Trim().Substring(0, 4) == "ovr_") &&
                                        !(l.Contains("new_") && l.Trim().Substring(0, 4) == "new_") &&
                                        !(l.Contains("hook_") && l.Trim().Substring(0, 5) == "hook_") &&
                                        !(l.Contains("repl_") && l.Trim().Substring(0, 5) == "repl_") &&
                                        !(l.Contains("nsub_") && l.Trim().Substring(0, 5) == "nsub_") &&
                                        !(l.Contains("/*overwrite") && l.Trim().Substring(0, 11) == "/*overwrite") &&
                                        !(l.Contains("overwrite*/") && l.Trim().Substring(0, 11) == "overwrite*/"));
                            }

                            // finished copying - compile insert
                            compileOverwrite(codeAddr, patchDir.Name + "/" + tmpDir.Name + "/" + ovrDir.Name);
                            insertOverwrite(codeAddr, ovId);

                            // cleanup and repeat for next block/file
                            System.IO.File.Delete(destFile);
                        }
                    }
                }
            }
        }

        public string parseLabel(string l)
        {
            string rem = l;
            while (rem.Contains("new_"))
            {
                // get "new_" label
                string sym = rem.Remove(0, rem.IndexOf("new_"));
                if (sym.Contains(" "))
                    sym = sym.Substring(0, sym.IndexOf(" "));
                if (sym.Contains("\t"))
                    sym = sym.Substring(0, sym.IndexOf("\t"));
                if (sym.Contains(","))
                    sym = sym.Substring(0, sym.IndexOf(","));
                if (sym.Contains("@"))
                    sym = sym.Substring(0, sym.IndexOf("@"));
                if (sym.Contains("/*"))
                    sym = sym.Substring(0, sym.IndexOf("/*"));
                if (sym.Contains("."))
                    sym = sym.Substring(0, sym.IndexOf("."));
                if (sym.Contains("("))
                    sym = sym.Substring(0, sym.IndexOf("("));

                // search "newcode.sym" for label address and insert
                string symAddr = "";
                FileInfo f = new FileInfo(tmpDir.FullName + "/newcode.sym");
                StreamReader s = f.OpenText();
                while (!s.EndOfStream)
                {
                    string sl = s.ReadLine();
                    if (sl.Contains(sym) && (sl.IndexOf(sym) + sym.Length) == sl.Length)
                    {
                        symAddr = sl.Substring(0, 8);
                        break;
                    }
                }
                s.Close();

                // skip over any unfound "new_" labels, let compiler deal with it
                if (symAddr == "") 
                {
                    int ind = l.IndexOf(sym) + sym.Length;
                    rem = l.Substring(ind, l.Length - ind);
                }
                // replace with address and repeat for any remaining "new_" labels
                else
                {
                    l = l.Remove(l.IndexOf(sym), sym.Length).Insert(l.IndexOf(sym), "0x" + symAddr);    // only replace first instance, to allow proper substitution of labels with similar names on same line (eg, "new_label" and "new_label2")
                    rem = l;
                }
            }

            return l;
        }

        public void compileOverwrite(int codeAddr, string compileDir)
        {
            PatchCompiler.compilePatch((uint)codeAddr, romdir, compileDir);
        }

        public void insertOverwrite(int ovrAddr, int ovId)
        {
            FileInfo f = new FileInfo(romdir.FullName + "/newcode.bin");
            if (!f.Exists) return;
            FileStream fs = f.OpenRead();

            byte[] newdata = new byte[fs.Length];
            fs.Read(newdata, 0, (int)fs.Length);
            fs.Close();

            for (int i = 0; i < newdata.Length; i++)
            {
                uint val = (uint)newdata[i];
                handler.writeToRamAddr(ovrAddr + i, val, ovId, 1);
            }
        }
        //
        

        public void generatePatch()
        {
        	int codeAddr = (int) getCodeAddr();
            Console.Out.WriteLine(String.Format("New code address: {0:X8}", codeAddr));

            FileInfo f = new FileInfo(romdir.FullName + "/newcode.bin");
            if (!f.Exists) return;
            FileStream fs = f.OpenRead();

            byte[] newdata = new byte[fs.Length];
            fs.Read(newdata, 0, (int)fs.Length);
            fs.Close();

            ByteArrayOutputStream extradata = new ByteArrayOutputStream();

            extradata.write(newdata);
            extradata.align(4);
            int hookAddr = codeAddr + extradata.getPos();


            f = new FileInfo(romdir.FullName + "/newcode.sym");
            StreamReader s = f.OpenText();

            while (!s.EndOfStream)
            {
                string l = s.ReadLine();

                int ind = -1;
                if (l.Contains("nsub_"))
                    ind = l.IndexOf("nsub_");
                if (l.Contains("hook_"))
                    ind = l.IndexOf("hook_");
                if (l.Contains("repl_"))
                    ind = l.IndexOf("repl_");

                if (ind != -1)
                {
                    int destRamAddr= parseHex(l.Substring(0, 8));    //Redirect dest addr
                    int ramAddr = parseHex(l.Substring(ind + 5, 8)); //Patched addr
                    uint val = 0;

                    int ovId = -1;
                    if (l.Contains("_ov_"))
                        ovId = parseHex(l.Substring(l.IndexOf("_ov_") + 4, 2));

                    int patchCategory = 0;

                    string cmd = l.Substring(ind, 4);
                    int thisHookAddr = 0;

                    switch(cmd)
                    {
                        case "nsub":
                            val = makeBranchOpcode(ramAddr, destRamAddr, false);
                            break;
                        case "repl":
                            val = makeBranchOpcode(ramAddr, destRamAddr, true);
                            break;
                        case "hook":
                            //Jump to the hook addr
                            thisHookAddr = hookAddr;
                            val = makeBranchOpcode(ramAddr, hookAddr, false);

                            uint originalOpcode = handler.readFromRamAddr(ramAddr, ovId);
                            
                            //TODO: Parse and fix original opcode in case of BL instructions
                            //so it's possible to hook over them too.
                            extradata.writeUInt(originalOpcode);
                            hookAddr += 4;
                            extradata.writeUInt(0xE92D5FFF); //push {r0-r12, r14}
                            hookAddr += 4;
                            extradata.writeUInt(makeBranchOpcode(hookAddr, destRamAddr, true));
                            hookAddr += 4;
                            extradata.writeUInt(0xE8BD5FFF); //pop {r0-r12, r14}
                            hookAddr += 4;
                            extradata.writeUInt(makeBranchOpcode(hookAddr, ramAddr+4, false));
                            hookAddr += 4;
                            extradata.writeUInt(0x12345678);
                            hookAddr += 4;
                            break;
                        default:
                            continue;
                    }

                    //Console.Out.WriteLine(String.Format("{0:X8}:{1:X8} = {2:X8}", patchCategory, ramAddr, val));
                    Console.Out.WriteLine(String.Format("              {0:X8} {1:X8}", destRamAddr, thisHookAddr));

                    handler.writeToRamAddr(ramAddr, val, ovId);
                }
            }

            s.Close();

            int newArenaOffs = codeAddr + extradata.getPos();
            handler.writeToRamAddr(ArenaLoOffs, (uint)newArenaOffs, -1);

            handler.sections.Add(new Arm9BinSection(extradata.getArray(), codeAddr, 0));
            // handler.saveSections();  //DY: moved to end CompilePatch()
        }

        private void loadArenaLoOffsFile(DirectoryInfo romdir)
        {
            FileInfo f = new FileInfo(romdir.FullName + "/arenaoffs.txt");
            StreamReader s = f.OpenText();
            string l = s.ReadLine();
            ArenaLoOffs = int.Parse(l, System.Globalization.NumberStyles.HexNumber);
            s.Close();
        }


        public static uint makeBranchOpcode(int srcAddr, int destAddr, bool withLink)
        {
            unchecked
            {
                uint res = (uint)0xEA000000;

                if (withLink)
                    res |= 0x01000000;

                int offs = (destAddr / 4) - (srcAddr / 4) - 2;
                offs &= 0x00FFFFFF;
                res |= (uint)offs;

                return res;
            }
        }


        public static uint parseUHex(string s)
        {
            return uint.Parse(s, System.Globalization.NumberStyles.HexNumber);
        }

        public static int parseHex(string s)
        {
            return int.Parse(s, System.Globalization.NumberStyles.HexNumber);
        }
    }
}
